/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Family of models for the MOESI cache coherence protocols:
//           (1) MSI   (Modified,                   Shared, Invalid)
//           (2) MOSI  (Modified, Owned,            Shared, Invalid)
//           (3) MESI  (Modified,        Exclusive, Shared, Invalid)
//           (4) MOESI (Modified, Owned, Exclusive, Shared, Invalid)
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/fsm/fsm.hpp>

#include <ostream>

using namespace cpptesk::hw;
using namespace cpptesk::hw::fsm;

namespace cpptesk {
namespace examples {
namespace moesi {

///////////////////////////////////////////////////////////////////////////////
// Cache states, stimuli and protocols
///////////////////////////////////////////////////////////////////////////////

/**
 * Cache line states.
 */
enum CacheState
{
    /// Modified (M).
    MODIFIED,
    /// Owned (O).
    OWNED,
    /// Exclusive (E).
    EXCLUSIVE,
    /// Shared (S).
    SHARED,
    /// Invalid (I).
    INVALID
};

inline std::ostream& operator << (std::ostream &out, CacheState state)
{
    switch(state)
    {
    case MODIFIED:  { return out << "M"; }
    case OWNED:     { return out << "O"; }
    case EXCLUSIVE: { return out << "E"; }
    case SHARED:    { return out << "S"; }
    case INVALID:   { return out << "I"; }
    default:        { return out << "-"; }
    }
}

/**
 * Cache stimuli.
 */
enum CacheStimulus
{
    ///////////////////////////////////////////////////////////////////////////
    // Local stimuli
    ///////////////////////////////////////////////////////////////////////////

    /// Read (R).
    READ,
    /// Exclusive Read (xR).
    READX,
    /// Write (W).
    WRITE,
    /// Evict (E).
    EVICT,

    ///////////////////////////////////////////////////////////////////////////
    // Remote stimuli
    ///////////////////////////////////////////////////////////////////////////

    /// Snoop Invalidate (sI).
    SN_I,
    /// Snoop Read (sR).
    SN_R
};

inline std::ostream& operator << (std::ostream &out, CacheStimulus stimulus)
{
    switch(stimulus)
    {
    case READ:  { return out << "R";  }
    case READX: { return out << "xR"; }
    case WRITE: { return out << "W";  }
    case EVICT: { return out << "E";  }
    case SN_I:  { return out << "sI"; }
    case SN_R:  { return out << "sR"; }
    default:    { return out << "-";  }
    }
}

/**
 * Cache protocols.
 */
enum CacheProtocol
{
    /// MSI cache coherence protocol.
    MSI,
    /// MOSI cache coherence protocol.
    MOSI,
    /// MESI cache coherence protocol.
    MESI,
    /// MOESI cache coherence protocol.
    MOESI
};

inline std::ostream& operator << (std::ostream &out, CacheProtocol protocol)
{
    switch(protocol)
    {
    case MSI:   { return out << "MSI";   }
    case MOSI:  { return out << "MOSI";  }
    case MESI:  { return out << "MESI";  }
    case MOESI: { return out << "MOESI"; }
    default:    { return out << "-";     }
    }
}

/// Checks whether the protocol has the Modified state.
inline bool has_modified(CacheProtocol)
{
    return true;
}

/// Checks whether the protocol has the Owned state.
inline bool has_owned(CacheProtocol protocol)
{
    return protocol == MOSI || protocol == MOESI;
}

/// Checks whether the protocol has the Exclusive state.
inline bool has_exclusive(CacheProtocol protocol)
{
    return protocol == MESI || protocol == MOESI;
}

/// Checks whether the protocol has the Shared state.
inline bool has_shared(CacheProtocol)
{
    return true;
}

/// Checks whether the protocol has the Invalid state.
inline bool has_invalid(CacheProtocol)
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Base cache coherence protocol
///////////////////////////////////////////////////////////////////////////////
class Core;

/**
 * Basic class for cache coherence protocols.
 */
CPPTESK_FSM(Protocol)
{
public:
    /// Factory of cache coherence protocols.
    static Protocol* create(CacheProtocol protocol, Core &core);

protected:
    /// Cache coherence protocol operates with the entire system.
    Protocol(Core &core):
        _core(core) {}

public:
    /// Basic transition actions.
    CPPTESK_DECLARE_ACTION(empty);
    CPPTESK_DECLARE_ACTION(error);
    CPPTESK_DECLARE_ACTION(read);
    CPPTESK_DECLARE_ACTION(write);
    CPPTESK_DECLARE_ACTION(evict);
    CPPTESK_DECLARE_ACTION(snoop);

private:
    /// Processor core.
    Core &_core;
};

///////////////////////////////////////////////////////////////////////////////
// MSI cache coherence protocol
///////////////////////////////////////////////////////////////////////////////

/**
 * FSM model of the MSI cache coherence protocol.
 */
class MsiProtocol: public Protocol
{
public:
    /// MSI transition table.
    MsiProtocol(Core &core): Protocol(core)
    {
        //=====================================================================
        CPPTESK_INITIAL_STATE ( INVALID                                       );
        //=====================================================================
        CPPTESK_TRANSITION    ( INVALID,   READ,   Protocol::read,   SHARED   );
        CPPTESK_TRANSITION    ( INVALID,   READX,  Protocol::read,   SHARED   );
        CPPTESK_TRANSITION    ( INVALID,   WRITE,  Protocol::write,  MODIFIED );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    READ,   Protocol::read,   SHARED   );
        CPPTESK_TRANSITION    ( SHARED,    WRITE,  Protocol::write,  MODIFIED );
        CPPTESK_TRANSITION    ( SHARED,    EVICT,  Protocol::evict,  INVALID  );
        //=====================================================================
        CPPTESK_TRANSITION    ( MODIFIED,  READ,   Protocol::read,   MODIFIED );
        CPPTESK_TRANSITION    ( MODIFIED,  WRITE,  Protocol::write,  MODIFIED );
        CPPTESK_TRANSITION    ( MODIFIED,  EVICT,  Protocol::evict,  INVALID  );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    SN_I,   Protocol::snoop,  INVALID  );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_I,   Protocol::snoop,  INVALID  );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_R,   Protocol::snoop,  SHARED   );
        //=====================================================================
    }
};

///////////////////////////////////////////////////////////////////////////////
// MOSI cache coherence protocol
///////////////////////////////////////////////////////////////////////////////

/**
 * FSM model of the MOSI cache coherence protocol.
 */
class MosiProtocol: public Protocol
{
public:
    /// MOSI transition table.
    MosiProtocol(Core &core): Protocol(core)
    {
        //=====================================================================
        CPPTESK_INITIAL_STATE ( INVALID                                       );
        //=====================================================================
        CPPTESK_TRANSITION    ( INVALID,   READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( INVALID,   READX,  Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( INVALID,   WRITE,  Protocol::write, MODIFIED  );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( SHARED,    WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( SHARED,    EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( OWNED,     READ,   Protocol::read,  OWNED     );
        CPPTESK_TRANSITION    ( OWNED,     WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( OWNED,     EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( MODIFIED,  READ,   Protocol::read,  MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( OWNED,     SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( OWNED,     SN_R,   Protocol::snoop, OWNED     );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_R,   Protocol::snoop, OWNED     );
        //=====================================================================
    }
};

///////////////////////////////////////////////////////////////////////////////
// MESI cache coherence protocol (aka Illinois protocol)
///////////////////////////////////////////////////////////////////////////////

/**
 * FSM model of the MSI cache coherence protocol.
 */
class MesiProtocol: public Protocol
{
public:
    /// MESI transition table.
    MesiProtocol(Core &core): Protocol(core)
    {
        //=====================================================================
        CPPTESK_INITIAL_STATE ( INVALID                                       );
        //=====================================================================
        CPPTESK_TRANSITION    ( INVALID,   READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( INVALID,   READX,  Protocol::read,  EXCLUSIVE );
        CPPTESK_TRANSITION    ( INVALID,   WRITE,  Protocol::write, MODIFIED  );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( SHARED,    WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( SHARED,    EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( EXCLUSIVE, READ,   Protocol::read,  EXCLUSIVE );
        CPPTESK_TRANSITION    ( EXCLUSIVE, WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( EXCLUSIVE, EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( MODIFIED,  READ,   Protocol::read,  MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( EXCLUSIVE, SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( EXCLUSIVE, SN_R,   Protocol::snoop, SHARED    );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_R,   Protocol::snoop, SHARED    );
        //=====================================================================
    }
};

///////////////////////////////////////////////////////////////////////////////
// MOESI cache coherence protocol
///////////////////////////////////////////////////////////////////////////////

/**
 * FSM model of the MOESI cache coherence protocol.
 */
class MoesiProtocol: public Protocol
{
public:
    /// MOESI transition table.
    MoesiProtocol(Core &core): Protocol(core)
    {
        //=====================================================================
        CPPTESK_INITIAL_STATE ( INVALID                                       );
        //=====================================================================
        CPPTESK_TRANSITION    ( INVALID,   READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( INVALID,   READX,  Protocol::read,  EXCLUSIVE );
        CPPTESK_TRANSITION    ( INVALID,   WRITE,  Protocol::write, MODIFIED  );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    READ,   Protocol::read,  SHARED    );
        CPPTESK_TRANSITION    ( SHARED,    WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( SHARED,    EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( EXCLUSIVE, READ,   Protocol::read,  EXCLUSIVE );
        CPPTESK_TRANSITION    ( EXCLUSIVE, WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( EXCLUSIVE, EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( OWNED,     READ,   Protocol::read,  OWNED     );
        CPPTESK_TRANSITION    ( OWNED,     WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( OWNED,     EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( MODIFIED,  READ,   Protocol::read,  MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  WRITE,  Protocol::write, MODIFIED  );
        CPPTESK_TRANSITION    ( MODIFIED,  EVICT,  Protocol::evict, INVALID   );
        //=====================================================================
        CPPTESK_TRANSITION    ( SHARED,    SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( EXCLUSIVE, SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( OWNED,     SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_I,   Protocol::snoop, INVALID   );
        CPPTESK_TRANSITION    ( EXCLUSIVE, SN_R,   Protocol::snoop, SHARED    );
        CPPTESK_TRANSITION    ( OWNED,     SN_R,   Protocol::snoop, OWNED     );
        CPPTESK_TRANSITION    ( MODIFIED,  SN_R,   Protocol::snoop, OWNED     );
        //=====================================================================
    }
};

}}} // cpptesk::examples::moesi

