/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Generate Diagnostic Report
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <fstream>

#include <hw/event.hpp>
#include <hw/sim_info.hpp>

#include <hw/report/model.hpp>

#include <hw/diagnostics/reaction_pair_vector.hpp>
#include <hw/diagnostics/reaction_unit_vector.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

class ReportGenerator
{
public:
    ReportGenerator(ReactionUnitVector &ruv, SimulationInfo &sim_info);
    virtual ~ReportGenerator();

    cpptesk::hw::report::Report run();
    
protected:
    ReactionUnitVector &_reaction_units;
    ReactionPairVector _reaction_pairs;

    cpptesk::hw::report::Report _report;
    cpptesk::hw::report::Failure _failure;
    
    SimulationInfo &_sim_info;

    unsigned _msg_size;
    
    /// Missing Reaction Timeout
    cycle_t _reaction_timeout;

    /// Unexpected Reaction Timeout
    cycle_t _unexpected_reaction_timeout;

    void _prepare_reaction_pairs();
    
    /// Number of expected MISSING reaction
    unsigned _expected_missing;
    /// Calculates the number of Expected MISSING in Reaction Vector
    uint _get_expected_missing();

    /// Number of optional MISSING reaction
    unsigned _optional_missing;
    /// Calculates the number of Optional MISSING in Reaction Vector
    uint _get_optional_missing();

    /// Add to the report the results of the test
    void _add_result(unsigned stimuli, unsigned reactions, unsigned reaction_normal, unsigned reaction_incorrect, 
                     unsigned reaction_missing, unsigned reaction_unexpected, uint64_t cycles, time_t time);
    /// Create statistic report
    void _add_statistic(unsigned stimuli, unsigned reactions, unsigned reaction_normal, unsigned reaction_incorrect, 
                        unsigned reaction_missing, unsigned reaction_unexpected);
    /// Create simulation report 
    void _add_simulation(int cycles, time_t time);

    /// Add fields of reaction to the failure
    void _add_fields(bool spec_get, bool impl_get, Reaction comp_reaction, bool error);

    /// Add item to fields
    void _add_fields(cpptesk::hw::report::FieldList &field_list, std::string field_string);

    /// 
    int _get_number_of_this_failure(bool collapsed, int count, Reaction main_reaction, int error_type_one, int error_type_two);

    /// Add errors in report
    void _add_error(bool collapsed, int &error_num, int iterator, int error_type_one);
    void _add_error(bool collapsed, int &error_num, int errors_type, int error_type_one, int error_type_two);
    
    void _add_error_group(int &error_num, ReactionPair example_pair, const int group);
    
    void _add_correct_data(Reaction comp_reaction);
    void _add_incorrect_data_error(Reaction comp_reaction);
    
    std::string _get_time_interval(Reaction reaction);
    std::string _get_time_interval(Reaction reaction, const ::cpptesk::hw::Interface *iface);
    std::string _get_iface_array_name(const ::cpptesk::hw::Interface *iface);

    void _add_title_error(int errors_type, const cpptesk::hw::Interface *iface_impl, 
                          Reaction &impl_reaction, error_type_t global_error_type, bool collapsed);
};

}}}
