/*
 * Copyright 2010-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

//==================================================================================================
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Message model
//==================================================================================================

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/clone.hpp>
#include <hw/errors.hpp>
#include <hw/field.hpp>
#include <hw/priority.hpp>
#include <hw/random.hpp>
#include <hw/time.hpp>
#include <hw/unused.hpp>

#include <stdint.h>
#include <string.h>

#include <sstream>
#include <ostream>
#include <string>
#include <vector>

//==================================================================================================
// Declaration of message classes
//==================================================================================================

/** @def CPPTESK_MESSAGE
 * Declares a model message.
 */
#define CPPTESK_MESSAGE(UserMessage) \
  struct UserMessage: public ::cpptesk::hw::Message

#ifdef HW_SHORT_NAMES
  #define MESSAGE(UserMessage) CPPTESK_MESSAGE(UserMessage)
#endif // HW_SHORT_NAMES

//==================================================================================================
// Declaration of message fields
//==================================================================================================

// IMPORTANT: To avoid warnings 1ull << 64 should be changed (1ull << 63) << 1.
#define __CPPTESK_FIELD_MASK(field_width) \
  (((1ull << ((field_width) - 1)) << 1) - 1)

/** @def CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY
 * Declares a masked typed field array.
 */
#define CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY(field_type, array_name, array_size, field_width, field_mask) \
  ::cpptesk::hw::Field<field_type, field_width, field_mask, array_size> array_name; \
  inline uint64_t get_##array_name(size_t i = 0) const { return array_name.get(i); } \
  inline void set_##array_name(uint64_t value, size_t i = 0) { array_name.set(value, i); }

/** @def CPPTESK_DECLARE_TYPED_FIELD_ARRAY
 * Declares a typed field array.
 */
#define CPPTESK_DECLARE_TYPED_FIELD_ARRAY(field_type, array_name, array_size, field_width) \
  CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY(field_type, array_name, array_size, field_width, __CPPTESK_FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_MASKED_TYPED_FIELD
 * Declares a masked typed field.
 */
#define CPPTESK_DECLARE_MASKED_TYPED_FIELD(field_type, field_name, field_width, field_mask) \
  CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY(field_type, field_name, 1, field_width, field_mask)

/** @def CPPTESK_DECLARE_TYPED_FIELD
 * Declares a typed field.
 */
#define CPPTESK_DECLARE_TYPED_FIELD(field_type, field_name, field_width) \
  CPPTESK_DECLARE_MASKED_TYPED_FIELD(field_type, field_name, field_width, __CPPTESK_FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_MASKED_FIELD_ARRAY
 * Declares a masked field array.
 */
#define CPPTESK_DECLARE_MASKED_FIELD_ARRAY(array_name, array_size, field_width, field_mask) \
  CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY(uint64_t, array_name, array_size, field_width, field_mask)

/** @def CPPTESK_DECLARE_FIELD_ARRAY
 * Declares a field array.
 */
#define CPPTESK_DECLARE_FIELD_ARRAY(array_name, array_size, field_width) \
  CPPTESK_DECLARE_MASKED_FIELD_ARRAY(array_name, array_size, field_width, __CPPTESK_FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_BIT_ARRAY
 * Declares a bit array.
 */
#define CPPTESK_DECLARE_BIT_ARRAY(array_name, array_size) \
  CPPTESK_DECLARE_MASKED_FIELD_ARRAY(array_name, array_size, 1)

/** @def CPPTESK_DECLARE_MASKED_FIELD
 * Declares a single masked field.
 */
#define CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, field_mask) \
  CPPTESK_DECLARE_MASKED_FIELD_ARRAY(field_name, 1, field_width, field_mask)

/** @def CPPTESK_DECLARE_FIELD
 * Declares a single field.
 */
#define CPPTESK_DECLARE_FIELD(field_name, field_width) \
  CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, __CPPTESK_FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_BIT
 * Declares a single one-bit field.
 */
#define CPPTESK_DECLARE_BIT(field_name) \
  CPPTESK_DECLARE_FIELD(field_name, 1)

#ifdef HW_SHORT_NAMES
  #define DECLARE_MASKED_TYPED_FIELD_ARRAY CPPTESK_DECLARE_MASKED_TYPED_FIELD_ARRAY
  #define DECLARE_TYPED_FIELD_ARRAY        CPPTESK_DECLARE_TYPED_FIELD_ARRAY
  #define DECLARE_MASKED_TYPED_FIELD       CPPTESK_DECLARE_MASKED_TYPED_FIELD
  #define DECLARE_TYPED_FIELD              CPPTESK_DECLARE_TYPED_FIELD
  #define DECLARE_MASKED_FIELD_ARRAY       CPPTESK_DECLARE_MASKED_FIELD_ARRAY
  #define DECLARE_FIELD_ARRAY              CPPTESK_DECLARE_FIELD_ARRAY
  #define DECLARE_BIT_ARRAY                CPPTESK_DECLARE_BIT_ARRAY
  #define DECLARE_MASKED_FIELD             CPPTESK_DECLARE_MASKED_FIELD
  #define DECLARE_FIELD                    CPPTESK_DECLARE_FIELD
  #define DECLARE_BIT                      CPPTESK_DECLARE_BIT
#endif // HW_SHORT_NAMES

//==================================================================================================
// Declaration and definition of a comparator
//==================================================================================================

#if defined(CPPTESKHW_VERSION) && ((CPPTESKHW_VERSION) >= 110428)

  /** @def CPPTESK_DECLARE_COMPARATOR
    * Declares a comparator in a message class.
    */
  #define CPPTESK_DECLARE_COMPARATOR() \
      virtual std::string compare(const ::cpptesk::hw::Message &message) const

  /** @def CPPTESK_DEFINE_COMPARATOR
    * Defines a comparator.
    */
  #define CPPTESK_DEFINE_COMPARATOR(UserMessage) \
      std::string UserMessage::compare(const ::cpptesk::hw::Message &message) const

  #ifdef HW_SHORT_NAMES
      #define DECLARE_COMPARATOR() CPPTESK_DECLARE_COMPARATOR()
      #define DEFINE_COMPARATOR(UserMessage)  CPPTESK_DEFINE_COMPARATOR(UserMessage)
  #endif // HW_SHORT_NAMES

#else

  /** @def CPPTESK_DECLARE_COMPARATOR
    * Declares a comparator in a message class.
    */
  #define CPPTESK_DECLARE_COMPARATOR(UserMessage) \
      /* Overloads the basic comparator */ \
      virtual std::string compare(const ::cpptesk::hw::Message &message) const \
          { return compare(dynamic_cast<const UserMessage&>(message)); } \
      virtual std::string compare(const UserMessage &message) const

  /** @def CPPTESK_DEFINE_COMPARATOR
    * Defines a comparator.
    */
  #define CPPTESK_DEFINE_COMPARATOR(UserMessage) \
      std::string UserMessage::compare(const UserMessage &message) const

  #ifdef HW_SHORT_NAMES
      #define DECLARE_COMPARATOR(UserMessage) CPPTESK_DECLARE_COMPARATOR(UserMessage)
      #define DEFINE_COMPARATOR(UserMessage)  CPPTESK_DEFINE_COMPARATOR(UserMessage)
  #endif // HW_SHORT_NAMES

#endif // CPPTESKHW_VERSION

/** @def CPPTESK_DECLARE_RANDOMIZER
 * Declares a randomizer in a message class.
 */
#define CPPTESK_DECLARE_RANDOMIZER() \
  virtual void randomize()

/** @def CPPTESK_DEFINE_RANDOMIZER
 * Defines a randomizer.
 */
#define CPPTESK_DEFINE_RANDOMIZER(UserMessage) \
  void UserMessage::randomize()

#ifdef HW_SHORT_NAMES
  #define DECLARE_RANDOMIZER()            CPPTESK_DECLARE_RANDOMIZER()
  #define DEFINE_RANDOMIZER(UserMessage)  CPPTESK_DEFINE_RANDOMIZER(UserMessage)
#endif // HW_SHORT_NAMES

//==================================================================================================
// Requirements traceability
//==================================================================================================

/** @def CPPTESK_SET_REQUIREMENT
 * Sets a requirement (some string explaning a reason for sending or receiving a message).
 */
#define CPPTESK_SET_REQUIREMENT(msg, req) \
  do { \
    std::stringstream _cpptesk_buffer; \
    _cpptesk_buffer << req; \
    _cpptesk_set_explanation(msg, __FILE__, __LINE__, _cpptesk_buffer.str()); \
  } while(false)

/** @def CPPTESK_SET_REQUIREMENT
 * Gets a message's requirement.
 */
#define CPPTESK_GET_REQUIREMENT(msg) \
  ((msg).getExplanation())

/** @def CPPTESK_SET_COMMENT
 * Alias of CPPTESK_SET_REQUIREMENT
 */
#define CPPTESK_SET_COMMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)

/** @def CPPTESK_GET_COMMENT
 * Alias of CPPTESK_GET_REQUIREMENT
 */
#define CPPTESK_GET_COMMENT(msg) CPPTESK_GET_REQUIREMENT(msg)

#ifdef HW_SHORT_NAMES
  #define SET_REQUIREMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)
  #define GET_REQUIREMENT(msg)      CPPTESK_GET_REQUIREMENT(msg)
  #define SET_COMMENT(msg, req)     CPPTESK_SET_COMMENT(msg, req)
  #define GET_COMMENT(msg)          CPPTESK_GET_COMMENT(msg)
#endif // HW_SHORT_NAMES

//==================================================================================================
// Field registration
//==================================================================================================

/** @def CPPTESK_ADD_FIELD
 * Registers a field in a message class.
 */
#define CPPTESK_ADD_FIELD(full_field_name) \
  registerField(#full_field_name, (Message::field_t)&full_field_name, true, true)

/** @def CPPTESK_ADD_INCOMPARABLE_FIELD
 * Registers an incomparable field in a message class.
 */
#define CPPTESK_ADD_INCOMPARABLE_FIELD(full_field_name) \
  registerField(#full_field_name, (Message::field_t)&full_field_name, false, true)

/** @def CPPTESK_ADD_NONRANDOMIZABLE_FIELD
 * Registers an nonrandomizable field in a message class.
 */
#define CPPTESK_ADD_NONRANDOMIZABLE_FIELD(full_field_name) \
  registerField(#full_field_name, (Message::field_t)&full_field_name, true, false)

/** @def CPPTESK_ADD_FIELD_ARRAY
 * Alias of CPPTESK_ADD_FIELD.
 */
#define CPPTESK_ADD_FIELD_ARRAY(full_array_name) \
  CPPTESK_ADD_FIELD(full_array_name)

/** @def CPPTESK_ADD_INCOMPARABLE_FIELD_ARRAY
 * Alias of CPPTESK_ADD_INCOMPARABLE_FIELD.
 */
#define CPPTESK_ADD_INCOMPARABLE_FIELD_ARRAY(full_array_name) \
  CPPTESK_ADD_INCOMPARABLE_FIELD(full_array_name)

/** @def CPPTESK_ADD_NONRANDOMIZABLE_FIELD_ARRAY
 * Alias of CPPTESK_ADD_NONRANDOMIZABLE_FIELD.
 */
#define CPPTESK_ADD_NONRANDOMIZABLE_FIELD_ARRAY(full_array_name) \
  CPPTESK_ADD_NONRANDOMIZABLE_FIELD(full_array_name)

#ifdef HW_SHORT_NAMES
  #define ADD_FIELD                       CPPTESK_ADD_FIELD
  #define ADD_INCOMPARABLE_FIELD          CPPTESK_ADD_INCOMPARABLE_FIELD
  #define ADD_NONRANDOMIZABLE_FIELD       CPPTESK_ADD_NONRANDOMIZABLE_FIELD
  #define ADD_FIELD_ARRAY                 CPPTESK_ADD_FIELD_ARRAY
  #define ADD_INCOMPARABLE_FIELD_ARRAY    CPPTESK_ADD_INCOMPARABLE_FIELD_ARRAY
  #define ADD_NONRANDOMIZABLE_FIELD_ARRAY CPPTESK_ADD_NONRANDOMIZABLE_FIELD_ARRAY
#endif // HW_SHORT_NAMES

//==================================================================================================
// Clone support
//==================================================================================================

#define CPPTESK_SUPPORT_CLONE(UserMessage) \
  virtual UserMessage* clone() const { return new UserMessage(*this); }

#ifdef HW_SHORT_NAMES
  #define SUPPORT_CLONE(UserMessage) CPPTESK_SUPPORT_CLONE(UserMessage)
#endif // HW_SHORT_NAMES

//==================================================================================================
// Message transmission delay & priority
//==================================================================================================

#define CPPTESK_MESSAGE_DELAY(msg, delay) \
  (msg).setDelay(delay)

#define CPPTESK_MESSAGE_LOCK(msg, pid) \
  (msg).setLock(pid)

#ifdef HW_SHORT_NAMES
  #define MESSAGE_DELAY(msg, delay) CPPTESK_MESSAGE_DELAY(msg, delay)
  #define MESSAGE_LOCK(msg, pid)    CPPTESK_MESSAGE_LOCK(msg, pid)
#endif // HW_SHORT_NAMES

//==================================================================================================
// Message randomization
//==================================================================================================

/** @def CPPTESK_RANDOMIZE_MESSAGE
 * Randomizes a message object.
 */
#define CPPTESK_RANDOMIZE_MESSAGE(msg) \
  (msg).randomize()

#ifdef HW_SHORT_NAMES
  #define RANDOMIZE_MESSAGE(msg) CPPTESK_RANDOMIZE_MESSAGE(msg)
#endif // HW_SHORT_NAMES

#define __CPPTESK_CLONE_MESSAGE(msg) \
  ((msg) ? (msg)->clone() : 0)

namespace cpptesk {
namespace hw {

typedef enum interlock_t {
  LOCK_NONE       = 0x00,
  LOCK_HOLD       = 0x01,
  LOCK_DELAY      = 0x02,
  LOCK_HOLD_DELAY = LOCK_HOLD | LOCK_DELAY
} interlock_t;

class Message: public Randomizable, public Cloneable {
public:
  /// Successfull comparison return value (empty string).
  const static std::string COMPARE_OK;

public:
    /// Comparison flags
  enum CompareFlags     {
    FIRST_MISMATCH_ONLY = 0x01
  };

  Message(): _valid(true), _explanation(""), _compare_flags(0),
      _unexpected(false), _optional(false), _id(0) {
    memset(&_sysdata, 0, sizeof(_sysdata));
  }

  virtual ~Message() {}

  Message(const Message &rhs);

  Message& operator = (const Message &rhs);

#ifdef CPPTESK_CLONE_INPUT_MESSAGE
  virtual Message* clone() const = 0;
#else
  virtual Message* clone() const { return this; }
#endif

  std::string getName() const { return _name; }
  void setName(const std::string &name) { _name = name; }

  void setSystemData(priority_t priority, interlock_t interlock, pid_t pid, cycle_t delay = 0) {
    _sysdata.priority  = priority;
    _sysdata.interlock = interlock;
    _sysdata.pid       = pid;
    _sysdata.delay     = delay;
  }

  void setPriority(priority_t priority) { _sysdata.priority = priority; }
  void setDelay(cycle_t delay)          { _sysdata.interlock = LOCK_DELAY; _sysdata.delay = delay; }
  void setHold(pid_t pid)               { _sysdata.interlock = LOCK_HOLD; _sysdata.pid = pid; }

  void setExplanation(const std::string &explanation) { _explanation = explanation; }
  const std::string& getExplanation() const { return _explanation; }

  priority_t getPriority() const { return _sysdata.priority; }
  interlock_t getInterlock() const { return _sysdata.interlock; }
  pid_t getPID() const { return _sysdata.pid; }
  int getDelay() const { return _sysdata.delay; }

  void setValid(bool valid = true) { _valid = valid; }
  void validate() { _valid = true; }
  void invalidate() { _valid = false; }
  bool isValid() const { return _valid; }

  void setValidFields(bool valid = true);
  void validateFields();
  void invalidateFields();

  bool isUnexpected() const { return _unexpected; }
  void setUnexpected(bool unexpected = true) { _unexpected = unexpected; }

  int getCompareFlags() const { return _compare_flags; }
  void setCompareFlags(int flags) { _compare_flags = flags; }

  bool isOptional() const { return _optional; }
  void setOptional(bool optional = true) { _optional = optional; }

  uint64_t getId() const { return _id; }
  void setId(uint64_t id) { _id = id; }

  /// Compares a message with another one and returns non-empty string if they are not equal.
  virtual std::string compare(const Message &message) const;
  virtual std::string forceCompare(const Message &message) const;

  /// Returns true if messages are equal (in terms of forceCompare()).
  bool equals(const Message &message) const { return forceCompare(message).empty(); }

  /// Returns a string representation of the message.
  virtual std::string toString() const;
  virtual std::string forceToString() const;

  /// Randomizes message data.
  virtual void randomize();

  /// Message invariant.
  virtual bool invariant() { return true; }

public:
  /// Registry of message fields.
  typedef Field<> Message::*field_t;
  typedef std::vector<field_t> fields_list_t;

  typedef fields_list_t::iterator iterator;
  typedef fields_list_t::const_iterator const_iterator;

  /// Fields iterator.
  iterator begin() { return _fields.begin(); }
  iterator end() { return _fields.end(); }

  const_iterator begin() const { return _fields.begin(); }
  const_iterator end() const { return _fields.end(); }

  size_t size() const { return _fields.size(); }

protected:
  void registerField(
      const std::string &full_name, field_t field, bool compare = true, bool randomize = true) {
    (this->*field).parseFullName(full_name);
    (this->*field).setComparable(compare);
    (this->*field).setRandomizable(randomize);

    _fields.push_back(field);

    setName((this->*field).getBaseName());
  }

  void copyFields(const Message &rhs);

protected:
  std::string _name;
  fields_list_t _fields;
  bool _valid;
  std::string _explanation;

  struct {
    priority_t priority;
    interlock_t interlock;
    pid_t pid;
    cycle_t delay;
  } _sysdata;

  int _compare_flags;
  bool _unexpected;
  bool _optional;
    uint64_t _id;
};

/**
 * Empty message being used as an input message for input-independent processes,
 * like reaction listeners and other infinite loops.
 */
CPPTESK_MESSAGE(EmptyMessage) {
public:
  EmptyMessage(): Message() {}
protected:
  static EmptyMessage* _instance;
public:
  static EmptyMessage& get() {
    if(!_instance) {
      _instance = new EmptyMessage();
    }
    return *_instance;
  }

  EmptyMessage(const EmptyMessage &rhs): Message(rhs) {}

  virtual ~EmptyMessage() {}

  CPPTESK_SUPPORT_CLONE(EmptyMessage)
};

#define CPPTESK_DEFAULT_MESSAGE ::cpptesk::hw::EmptyMessage::get()

inline std::ostream& operator << (std::ostream& out, const Message &msg) {
  return out << msg.toString();
}

// Internal function that is used in some defines.
inline Message& _cpptesk_set_explanation(
    Message &msg, const std::string &file, int line, const std::string &req = "") {
  if(req.empty()) {
    return msg;
  }

  std::stringstream buffer;
  buffer << msg.getExplanation();

  if(!msg.getExplanation().empty()) {
    buffer << ", ";
  }

  buffer << req;

#if 0
  // The information on file and line is kept in the process call descriptor
  buffer << " (" << file << ":" << line << ")";
#else
  CPPTESK_UNUSED_ARG(file);
  CPPTESK_UNUSED_ARG(line);
#endif

  msg.setExplanation(buffer.str());

  return msg;
}

template<class UserMessage>
inline UserMessage& _cpptesk_cast_message(
    Message &message, const std::string &file, int line) {
  UserMessage *casted_message = dynamic_cast<UserMessage*>(&message);
  __CPPTESK_ASSERTION_FILE_LINE(casted_message,
                                "Cannot not cast a message to the type given", file, line);
  return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_cast_message(
    const Message &message, const std::string &file, int line) {
  const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
  __CPPTESK_ASSERTION_FILE_LINE(casted_message,
                                "Cannot not cast a message to the type given", file, line);
  return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_const_cast_message(
    Message &message, const std::string &file, int line) {
  const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
  __CPPTESK_ASSERTION_FILE_LINE(casted_message,
                                "Cannot not cast a message to the type given", file, line);
  return *casted_message;
}

template<class UserMessage>
inline const UserMessage& _cpptesk_const_cast_message(
    const Message &message, const std::string &file, int line) {
  const UserMessage *casted_message = dynamic_cast<const UserMessage*>(&message);
  __CPPTESK_ASSERTION_FILE_LINE(casted_message,
                                "Cannot not cast a message to the type given", file, line);
  return *casted_message;
}

}} // namespace cpptesk::hw
