/*
 * Copyright 2010-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

//==================================================================================================
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Macros for declaring interfaces, default adapter, and arbiters
//==================================================================================================

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/debug.hpp>

#if defined(CPPTESKHW_VERSION) && ((CPPTESKHW_VERSION) >= 111010)

//==================================================================================================
// Default adapter naming
//==================================================================================================

/** @def CPPTESK_DEFAULT_INPUT_ADAPTER
 * Specifies the name of the default input adapter.
 */
#define CPPTESK_DEFAULT_INPUT_ADAPTER(iface) \
  /* IMPORTANT: 'iface' should start the default adapter name */ \
  iface##_serialize

/** @def CPPTESK_DEFAULT_OUTPUT_ADAPTER
 * Specifies the name of the default output adapter.
 */
#define CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface) \
  /* IMPORTANT: 'iface' should start the default adapter name */ \
  iface##_deserialize

//==================================================================================================
// Declaring/defining default adapters
//==================================================================================================

/** @def CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER
 * Declares the default input adapter.
 */
#define CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) \
  typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_input_##iface; \
  CPPTESK_DECLARE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(iface))

/** @def CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER
 * Declares the default output adapter.
 */
#define CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) \
  typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_output_##iface; \
  CPPTESK_DECLARE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface))

/** @def CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER
 * Defines the default input adapter.
 */
#define CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER(AdapterClass_iface) \
  CPPTESK_DEFINE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(AdapterClass_iface))

/** @def CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER
 * Defines the default output adapter.
 */
#define CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface) \
  CPPTESK_DEFINE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface))

//==================================================================================================
// Default interface messages
//==================================================================================================

/** @def CPPTESK_GET_IFACE_MESSAGE_TYPE
 * Specifies the name of the interface message type.
 */
#define CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) \
  MessageType_##iface

/** @def CPPTESK_GET_IFACE_MESSAGE
 * Specifies the name of the default interface message.
 */
#define CPPTESK_GET_IFACE_MESSAGE(iface) \
  /* IMPORTANT: 'iface' should end the message name to support interface arrays */ \
  message_##iface

//==================================================================================================
// Auxiliary macros for declaring interfaces
//==================================================================================================

#define __CPPTESK_DECLARE_IFACE(iface_class, iface, msg_type) \
  typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
  CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface); \
  iface_class iface

#define __CPPTESK_DECLARE_IFACE_ARRAY(iface_class, iface, msg_type, array_size) \
  typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
  CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[array_size]; \
  iface_class iface[array_size]

#define __CPPTESK_DECLARE_IFACE_ARRAY_2D(iface_class, iface, msg_type, i_size, j_size) \
  typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
  CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[i_size][j_size]; \
  iface_class iface[i_size][j_size]

//==================================================================================================
// Declaring general interfaces (no default adapters)
//==================================================================================================

/** @def CPPTESK_DECLARE_IFACE
 * Declares the general interface (no default adapter).
 */
#define CPPTESK_DECLARE_IFACE(iface, msg_type) \
  __CPPTESK_DECLARE_IFACE(::cpptesk::hw::Interface, iface, msg_type)

/** @def CPPTESK_DECLARE_IFACE_ARRAY
 * Declares the array of general interfaces (no default adapters).
 */
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, msg_type, array_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::Interface, iface, msg_type, array_size)

/** @def CPPTESK_DECLARE_IFACE_ARRAY_2D
 * Declares the array of the general interfaces (no default adapters).
 */
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, msg_type, i_size, j_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::Interface, iface, msg_type, i_size, j_size)

//==================================================================================================
// Declaring input interfaces (with the default input adapters)
//==================================================================================================

/** @def CPPTESK_DECLARE_INPUT
 * Declares the input interface (with the default input adapter).
 */
#define CPPTESK_DECLARE_INPUT(iface, msg_type) \
  __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InputInterface, iface, msg_type); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }

/** @def CPPTESK_DECLARE_INPUT_ARRAY
 * Declares the array of input interfaces (with the default input adapters).
 */
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, msg_type, array_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InputInterface, iface, msg_type, array_size); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }

/** @def CPPTESK_DECLARE_INPUT_ARRAY_2D
 * Declares the 2D array of input interfaces (with the default input adapters).
 */
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, msg_type, i_size, j_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InputInterface, iface, msg_type, i_size, j_size); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }

//==================================================================================================
// Declaring output interfaces (with the default output adapters)
//==================================================================================================

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares the output interface (with the default output adapter and reaction arbiter).
 */
#define CPPTESK_DECLARE_OUTPUT(iface, msg_type, arbiter_type) \
  __CPPTESK_DECLARE_IFACE(::cpptesk::hw::OutputInterface, iface, msg_type); \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);

/** @def CPPTESK_DECLARE_OUTPUT_ARRAY
 * Declares the array of output interfaces (with the default output adapters and reaction arbiters).
 */
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::OutputInterface, iface, msg_type, array_size); \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);

/** @def CPPTESK_DECLARE_OUTPUT_ARRAY_2D
 * Declares the 2D array of output interfaces (with the default output adapters and reaction arbiters).
 */
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::OutputInterface, iface, msg_type, i_size, j_size); \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

//==================================================================================================
// Declaring input/output interfaces (with the default input/output adapters)
//==================================================================================================

/** @def CPPTESK_DECLARE_INOUT
 * Declares the input/output interface (with the default input/output adapters and reaction arbiter).
 */
#define CPPTESK_DECLARE_INOUT(iface, msg_type, arbiter_type) \
  __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InoutInterface, iface, msg_type); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);

/** @def CPPTESK_DECLARE_INOUT_ARRAY
 * Declares the array of input/output interfaces (with the default input/output adapters and reaction arbiters).
 */
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InoutInterface, iface, msg_type, array_size); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);

/** @def CPPTESK_DECLARE_INOUT_ARRAY_2D
 * Declares the 2D array of input/output interfaces (with the default input/output adapters and reaction arbiters).
 */
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
  __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InoutInterface, iface, msg_type, i_size, j_size); \
  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
  CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

#else // CPPTESKHW_VERSION

//==================================================================================================
// Declaring general interfaces
//==================================================================================================

/** @def CPPTESK_DECLARE_IFACE
 * Declares the interface.
 */
#define CPPTESK_DECLARE_IFACE(iface) \
  ::cpptesk::hw::Interface iface

/** @def CPPTESK_DECLARE_IFACE_ARRAY
 * Declares the array of interfaces.
 */
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, array_size) \
  ::cpptesk::hw::Interface iface[array_size]

/** @def CPPTESK_DECLARE_IFACE_ARRAY_2D
 * Declares the 2D array of interfaces.
 */
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, i_size, j_size) \
  ::cpptesk::hw::Interface iface[i_size][j_size]

//==================================================================================================
// Declaring input interfaces
//==================================================================================================

/** @def CPPTESK_DECLARE_INPUT
 * Declares the input interface.
 */
#define CPPTESK_DECLARE_INPUT(iface) \
  ::cpptesk::hw::InputInterface iface

/** @def CPPTESK_DECLARE_INPUT_ARRAY
 * Declares the array of input interfaces.
 */
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, array_size) \
  ::cpptesk::hw::InputInterface iface[array_size]

/** @def CPPTESK_DECLARE_INPUT_ARRAY_2D
 * Declares the 2D array of input interfaces.
 */
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, i_size, j_size) \
  ::cpptesk::hw::InputInterface iface[i_size][j_size]

//==================================================================================================
// Declaring output interfaces
//==================================================================================================

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares the output interface.
 */
#define CPPTESK_DECLARE_OUTPUT(iface) \
  ::cpptesk::hw::OutputInterface iface

/** @def CPPTESK_DECLARE_OUTPUT_ARRAY
 * Declares the array of output interfaces.
 */
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, array_size) \
  ::cpptesk::hw::OutputInterface iface[array_size]

/** @def CPPTESK_DECLARE_OUTPUT_ARRAY_2D
 * Declares the 2D array of output interfaces.
 */
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
  ::cpptesk::hw::OutputInterface iface[i_size][j_size]

//==================================================================================================
// Declaring input/output interfaces
//==================================================================================================

/** @def CPPTESK_DECLARE_INOUT
 * Declares the input/output interface.
 */
#define CPPTESK_DECLARE_INOUT(iface) \
  ::cpptesk::hw::InoutInterface iface

/** @def CPPTESK_DECLARE_INOUT_ARRAY
 * Declares the array of input/output interfaces.
 */
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, array_size) \
  ::cpptesk::hw::InoutInterface iface[array_size]

/** @def CPPTESK_DECLARE_INOUT_ARRAY_2D
 * Declares the 2D array of input/output interfaces.
 */
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, i_size, j_size) \
  ::cpptesk::hw::InoutInterface iface[i_size][j_size]

#endif // CPPTESKHW_VERSION

//==================================================================================================
// Declaring reaction arbiters
//==================================================================================================

/** @def CPPTESK_DECLARE_ARBITER
 * Declares the reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_ARBITER(type, arbiter) \
  type arbiter

/** @def CPPTESK_DECLARE_ARBITER_ARRAY
 * Declares the array of reaction arbiters (matchers).
 */
#define CPPTESK_DECLARE_ARBITER_ARRAY(type, arbiter, array_size) \
  type arbiter[array_size]

/** @def CPPTESK_DECLARE_ARBITER_ARRAY_2D
 * Declares the 2D array of reaction arbiters (matchers).
 */
#define CPPTESK_DECLARE_ARBITER_ARRAY_2D(type, arbiter, i_size, j_size) \
  type arbiter[i_size][j_size]

//==================================================================================================
// Declaring two-level reaction arbiters
//==================================================================================================

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER
 * Declares the two-level reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER(A1, A2, arbiter) \
  ::cpptesk::hw::TwoLevelArbiter<A1, A2> arbiter

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY
 * Declares the array of two-level reaction arbiters (matchers).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY(A1, A2, arbiter, array_size) \
  ::cpptesk::hw::TwoLevelArbiter<A1, A2> arbiter[array_size]

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY_2D
 * Declares the 2D array of two-level reaction arbiters (matchers).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY_2D(A1, A2, arbiter, i_size, j_size) \
  ::cpptesk::hw::TwoLevelArbiter<A1, A2> arbiter[i_size][j_size]

//==================================================================================================
// Short names
//==================================================================================================

#ifdef HW_SHORT_NAMES
  #define GET_IFACE_MESSAGE_TYPE             CPPTESK_GET_IFACE_MESSAGE_TYPE
  #define GET_IFACE_MESSAGE                  CPPTESK_GET_IFACE_MESSAGE
  #define DEFAULT_INPUT_ADAPTER              CPPTESK_DEFAULT_INPUT_ADAPTER
  #define DEFAULT_OUTPUT_ADAPTER             CPPTESK_DEFAULT_OUTPUT_ADAPTER
  #define DECLARE_DEFAULT_INPUT_ADAPTER      CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER
  #define DECLARE_DEFAULT_OUTPUT_ADAPTER     CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER
  #define DEFINE_DEFAULT_INPUT_ADAPTER       CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER
  #define DEFINE_DEFAULT_OUTPUT_ADAPTER      CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER
  #define DECLARE_IFACE                      CPPTESK_DECLARE_IFACE
  #define DECLARE_IFACE_ARRAY                CPPTESK_DECLARE_IFACE_ARRAY
  #define DECLARE_IFACE_ARRAY_2D             CPPTESK_DECLARE_IFACE_ARRAY_2D
  #define DECLARE_INPUT                      CPPTESK_DECLARE_INPUT
  #define DECLARE_INPUT_ARRAY                CPPTESK_DECLARE_INPUT_ARRAY
  #define DECLARE_INPUT_ARRAY_2D             CPPTESK_DECLARE_INPUT_ARRAY_2D
  #define DECLARE_OUTPUT                     CPPTESK_DECLARE_OUTPUT
  #define DECLARE_OUTPUT_ARRAY               CPPTESK_DECLARE_OUTPUT_ARRAY
  #define DECLARE_OUTPUT_ARRAY_2D            CPPTESK_DECLARE_OUTPUT_ARRAY_2D
  #define DECLARE_INOUT                      CPPTESK_DECLARE_INOUT
  #define DECLARE_INOUT_ARRAY                CPPTESK_DECLARE_INOUT_ARRAY
  #define DECLARE_INOUT_ARRAY_2D             CPPTESK_DECLARE_INOUT_ARRAY_2D
  #define DECLARE_ARBITER                    CPPTESK_DECLARE_ARBITER
  #define DECLARE_ARBITER_ARRAY              CPPTESK_DECLARE_ARBITER_ARRAY
  #define DECLARE_ARBITER_ARRAY_2D           CPPTESK_DECLARE_ARBITER_ARRAY_2D
  #define DECLARE_TWO_LEVEL_ARBITER          CPPTESK_DECLARE_TWO_LEVEL_ARBITER
  #define DECLARE_TWO_LEVEL_ARBITER_ARRAY    CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY
  #define DECLARE_TWO_LEVEL_ARBITER_ARRAY_2D CPPTESK_DECLARE_TWO_LEVEL_ARBITER_ARRAY_2D
#endif // HW_SHORT_NAMES

