/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Random number generation utilities
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/random/generator.hpp>
#include <hw/utils/singleton.hpp>

namespace cpptesk {
namespace hw {

/**
 * Random number generation utilities.
 */
class Randomization: public Singleton<Randomization>
{
    friend class Singleton<Randomization>;

public:

    template<typename T> static inline T abs(T val)
    {
        return val & ~((T)1 << (8*sizeof(T) - 1));
    }

    template<typename T> static inline T mask(unsigned width)
    {
        return width == 64 ? 0xffffFFFFffffFFFFull : ((T)1 << width) - 1;
    }

    template<typename T> static inline T mask(unsigned lo, unsigned hi)
    {
        return mask<T>((hi - lo) + 1) << lo;
    }

    template<typename T> inline T rand()
    {
        uint64_t res = Generator::get().next();

        res &= mask<uint64_t>(8*sizeof(T));

        for(unsigned i = 1; i <= sizeof(T)/4; i++)
            { res <<= 32; res |= Generator::get().next(); }

        return (T)res;
    }

    template<typename T> inline T rand_range(T min, T max)
    {
        T base = abs<T>(rand<T>());

        return min + base % ((max - min) + 1);
    }

    template<typename T> inline T rand_field(unsigned width)
    {
        return rand<T>() & mask<T>(width);
    }

    template<typename T> inline T rand_field(unsigned lo, unsigned hi)
    {
        return rand<T>() & mask<T>(lo, hi);
    }

    template<typename T> inline T rand_choice(T array[], unsigned size)
    {
        CPPTESK_ASSERTION(size > 0, "rand_choice: array has zero size");
        return array[rand_range<unsigned>(0u, size - 1)];
    }

    template<typename T> inline T rand_excluding(T array[], unsigned size)
    {
        T result;
	do { result = rand<T>(); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand_range_excluding(T min, T max, T array[], unsigned size)
    {
        T result;
	do { result = rand_range<T>(min, max); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand_field_excluding(unsigned width, T array[], unsigned size)
    {
        T result;
        do { result = rand_field<T>(width); } while(contains<T>(array, size, result));
        return result;
    }

    template<typename T> inline T rand_field_excluding(unsigned lo, unsigned hi, T array[], unsigned size)
    {
        T result;
	do { result = rand_field<T>(lo, hi); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand(unsigned width)
    {
        return rand_field<T>(width);
    }

    template<typename T> inline T rand(unsigned lo, unsigned hi)
    {
        return rand_field<T>(lo, hi);
    }

    template<typename T> inline T rand_excluding(unsigned width, T array[], unsigned size)
    {
        return rand_field_excluding<T>(width, array, size);
    }

    template<typename T> inline T rand_excluding(unsigned lo, unsigned hi, T array[], unsigned size)
    {
        return rand_field_excluding<T>(lo, hi, array, size);
    }

    inline bool rand_indicator(float probability)
    {
        CPPTESK_ASSERTION(probability <= 1.0, "rand_indicator: probability is greater than 1.0");
        return rand_range<unsigned>(0, 99) < (unsigned)(100.0 * probability);
    }

protected:
    Randomization() {}
    virtual ~Randomization() {}

    /// Checks whether the array contains the given item.
    template<typename T> static bool contains(T array[], unsigned size, T item)
    {
        for(unsigned i = 0; i < size; i++)
        {
            if(item == array[i])
                { return true; }
        }

        return false;
    }
};

}} // namespace cpptesk::hw

