/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Class for Printers of Diagnostics Results in xml format
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/report/report_printer.hpp>

#include <hw/time.hpp>

#include <hw/diagnostics/reaction_pair_vector.hpp>

namespace cpptesk {
namespace hw {
namespace report {

struct TagParameter
{
    std::string name;
    std::string value;

    // For XML tag parameter: String
    TagParameter(std::string in_name, std::string in_value) { name = in_name; value = in_value; }
    
    // For XML tag parameter: Cycle
    TagParameter(std::string in_name, cycle_t cycle)
    {
        name = in_name;
        
        std::stringstream stream;
        stream << cycle;
        value = stream.str();
    }
    
    // For XML tag parameter: Boolean
    TagParameter(std::string in_name, bool status)
    {
        name = in_name;
        value = status?"true":"false";
    }
    
    // For XML tag parameter: Time
    TagParameter(std::string in_name, time_t time)
    {
        name = in_name;
        
        std::stringstream temp;
        temp << time;

        value = temp.str();
    }
    
    // For XML tag parameter: Double
    TagParameter(std::string in_name, double in_value)
    {
        name = in_name;
        
        std::stringstream temp;
        temp << in_value;

        value = temp.str();
    }
};

struct Tag
{
    std::string name;
    std::vector<Tag> tags;
    std::vector<TagParameter> params;
    
    void add_parameter(TagParameter param)
    {
        params.push_back(param);
    }
    
    void add_subtag(Tag tag)
    {
        tags.push_back(tag);
    }
    
    Tag(std::string tag_name) { name = tag_name; }

    std::string toString();
    std::string operator<<(Tag &tag);
};

class XMLPrinter: public ReportPrinter
{
public:
    XMLPrinter();
    virtual ~XMLPrinter();

    virtual void print(std::ostream &out, const Report &report);
    void print(std::string xml_trace_name, const Report &report);

private:
    // Print XML structure
    void _print(std::string str, uint tab);
    void _print_tag_border(std::string tag_name, bool left, uint tab);
    void _print_open_tag(std::string tag_name, uint tab);
    void _print_close_tag(std::string tag_name, uint tab);
    void _print_param(TagParameter param, uint tab);
    void _print_tag(Tag tag, uint tab);

    // Add to xml structure group of failures
    Tag _process_group(Tag &failures_tag, const cpptesk::hw::report::Failures &failures);
    
    // Add to XML structure one failure
    Tag _process_failure(const cpptesk::hw::report::Failure &failure);

    // Add to XML structure data of failure
    Tag _process_data(const cpptesk::hw::report::DataFailure &data);

    // Add to XML structure field of data
    Tag _process_field(const cpptesk::hw::report::FieldList &field_list, const std::string name);
    
    // Add to XML structure time of failure
    Tag _process_time(const cpptesk::hw::report::TimeFailure &time);
    
    // Add to XML structure iface of failure
    Tag _process_iface(const cpptesk::hw::report::IfaceFailure &iface);
    
    // Add to XML structure statistic information
    Tag _process_statistics(Tag &statistics_tag, const cpptesk::hw::report::Statistics &statistics);

    // Add to XML structure simulation information
    Tag _process_simulation(Tag &simulation_tag, const cpptesk::hw::report::Simulation &simulation);
    
    // Tab char for the start of the line
    const char *_tab;
    
    std::ofstream *_xml_trace;
    std::vector<std::string> _group_types;
};

}}} // cpptesk::hw::report
